#!/usr/bin/env python
# -*- coding: UTF-8 -*-
'''
    The MIT License (MIT)

    Copyright (c) 2015 Alfred Wassermann

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.
'''
helptext = '''
    This program reads q-Steiner systems 2-(13,3,1;2) from the file
    "qSteinerSystems.txt" and computes intersection properties
    of the blocks of the design.

    Each design is given by 15 orbit representatives under the group action of
    the group G = <singer, phi>, where phi is the Frobenius automorphism in
    GF(2^13). The order of the group is (2^13-1)*13 = 106483.

    The program is started with the command:
        python check_q_steiner.py <solution_number> <orbit_number>

    However, it is recommended to use pypy (http://pypy.org) instead of the
    standard python interpreter.

    It picks out the q-Steiner system with number <solution_number> and from this
    design the orbit representative of orbit with number <orbit_number>
    (0 <= <orbit_number> <= 14).
    Then this orbit representative B which is a subspace of dimension 3 is
    intersected with all subspaces of all orbits and the intersection spectrum
    is printed. There should be exactly one orbit which intersects B in dimension
    three, namely the block itself, all other intersetcion should have dimension
    zero or one.
    '''
import sys

'''Frobenius automorphism of order 13'''
phi = [[1,0,0,0,0,0,0,0,0,0,0,1,0],
    [0,0,0,0,0,0,0,1,0,0,0,1,1],
    [0,1,0,0,0,0,0,1,0,0,0,0,0],
    [0,0,0,0,0,0,0,0,1,0,0,1,1],
    [0,0,1,0,0,0,0,1,1,0,0,1,1],
    [0,0,0,0,0,0,0,1,0,1,0,0,0],
    [0,0,0,1,0,0,0,0,1,1,0,0,1],
    [0,0,0,0,0,0,0,0,1,0,1,0,0],
    [0,0,0,0,1,0,0,0,0,1,1,0,0],
    [0,0,0,0,0,0,0,0,0,1,0,1,0],
    [0,0,0,0,0,1,0,0,0,0,1,1,0],
    [0,0,0,0,0,0,0,0,0,0,1,0,1],
    [0,0,0,0,0,0,1,0,0,0,0,1,1]]

'''Generator of the Singer cycle, order 8191'''
singer =[[0,0,0,0,0,0,0,0,0,0,0,0,1],
    [1,0,0,0,0,0,0,0,0,0,0,0,1],
    [0,1,0,0,0,0,0,0,0,0,0,0,0],
    [0,0,1,0,0,0,0,0,0,0,0,0,1],
    [0,0,0,1,0,0,0,0,0,0,0,0,1],
    [0,0,0,0,1,0,0,0,0,0,0,0,0],
    [0,0,0,0,0,1,0,0,0,0,0,0,0],
    [0,0,0,0,0,0,1,0,0,0,0,0,0],
    [0,0,0,0,0,0,0,1,0,0,0,0,0],
    [0,0,0,0,0,0,0,0,1,0,0,0,0],
    [0,0,0,0,0,0,0,0,0,1,0,0,0],
    [0,0,0,0,0,0,0,0,0,0,1,0,0],
    [0,0,0,0,0,0,0,0,0,0,0,1,0]]

def MulMatVec(mat, vec, q = 2):
    '''Matrix-vector multiplication over GF(q), q prime

    Arguments:
    mat -- n x m matrix
    vec -- vector of length m
    q -- prime (default = 2)

    Return: mat * vec over GF(q)
    '''
    r = len(mat)
    c = len(mat[0])
    w = [0]*r

    if q == 2:
        for i in xrange(r):
            s = 0
            for j in xrange(c):
                s ^= mat[i][j] & vec[j]
            w[i] = s
    else:
        for i in xrange(r):
            s = 0
            #for j in xrange(c):
            #        s += (mat[i][j] * vec[j])
            for j, m in enumerate(mat[i]):
                s += (m * vec[j])
            w[i] = ((s % q) + q) % q

    return w

def Action(g, s, q = 2):
    '''Let the matrix group element g act on subspace s over GF(q), q prime

    Arguments:
    g -- n x n matrix, element of matrix group
    s -- k x n matrix, basis of k-subspace in GF(q)^n. The rows are the basis vectors

    Returns: s'^T = g * s^T
    '''
    k = len(s)
    w = [[]] * k
    for i in xrange(k):
        w[i] = MulMatVec(g, s[i], q)
    return w

def Orbit(v):
    '''Hard coded orbit algorithm for this specific group <Singer, phi>.
    Generator function, yields all orbit elements.

    Arguments:
    v -- orbit representative
    '''
    cnt = 0
    w = v
    for i in range(8191):
        for j in range(13):
            w = Action(phi, w)
            cnt += 1
            yield w

        w = Action(singer, w)

def ToReducedRowEchelonForm(mat, q = 2):
    '''Determine reduced row echelon form over GF(q) for q prime!
    Changes the matrix mat in place. Here, q is restricted to q = 2.
    This code is an adaption from http://rosettacode.org/wiki/Reduced_row_echelon_form.

    Arguments:
    mat -- (k x n) matrix
    q -- prime (default = 2)
    '''
    if not mat: return
    if q != 2:
        print >>sys.stderr, "q must be 2 in ToReducedRowEchelonForm()!"
        sys.exit(1)

    lead = 0
    rowCount = len(mat)
    columnCount = len(mat[0])
    rows = len(mat)
    cols = len(mat[0])

    for r in range(rows):
        if lead >= cols:
            return

        i = r
        while mat[i][lead] == 0:
            i += 1
            if i == rows:
                i = r
                lead += 1
                if cols == lead:
                    return

        mat[i], mat[r] = mat[r], mat[i]

        lv = mat[r][lead]
        for i in range(rows):
            if i != r:
                lv = mat[i][lead]
                if lv != 0:
                    mat[i] = [ (iv + rv) % 2 for rv, iv in zip(mat[r], mat[i])]
        lead += 1

def Rank(mat, q = 2):
    '''Rank of matrix over GF(q), q prime

    Arguments:
    mat: (k x n) matrix
    q -- prime (default = 2)

    Return integer: rank(mat) over GF(q)
    '''
    r = len(mat)
    rnk = r
    m = mat[:]
    ToReducedRowEchelonForm(m, q)
    for i in range(r-1, -1, -1):
        if sum(m[i]) == 0:
            rnk -= 1
        else:
            break
    return rnk

def PrintOrbitRepresentatives(solution):
    '''Print to stdout orbit representatives

    Arguments:
    solution -- Set of orbit representatives
    '''
    for i, s in enumerate(solution):
        print "Orbit %d:" % i, s

def IntersectionNumbers(solution, orbit_number):
    '''Given a set of orbit representatives, compute the intersection properties
    of the representative of orbit 'orbit_number' with all other orbits.
    Prints the result to stdout.

    Arguments:
    solution -- Set of orbit representatives
    orbit_number: orbit representative (given as orbit number) which is intersected
    with all orbits.
    '''
    # Select orbit representative for testing
    rep = solution[orbit_number]
    k = len(rep)

    intersection = [0] * (k + 1)
    # Run through all orbits and intersect rep with all other blocks
    for i, other_rep in enumerate(solution):
        orbitintersection = [0] * (k + 1)
        for space in Orbit(other_rep):
            #
            # The rank of the intersection is
            # Rank(rep) + Rank(space) -  Rank(rep + space)
            #
            rank = 3 + 3 - Rank(rep + space, q)
            orbitintersection[rank] += 1
        for j in range(k + 1):
            intersection[j] += orbitintersection[j]

        print "Intersection with orbit %d:" % i,
        for j in range(k + 1):
            print " #%d-dim = %d," % (j, orbitintersection[j]),
        print

    print "\nIntersection spectrum:",
    for j in range(k + 1):
        print " #%d-dim = %d," % (j, intersection[j]),
    print

if __name__ == '__main__':
    n = 13
    k = 3
    t = 2
    q = 2

    if len(sys.argv) < 3 or sys.argv[1] in ['-?', '-h', '--help']:
        print >>sys.stderr, helptext
        print >>sys.stderr, "    Call:"
        print >>sys.stderr, "        python check_q_steiner.py <solution_number> <orbit_number>"
        print >>sys.stderr, "    Example:"
        print >>sys.stderr, "        python check_q_steiner.py 0 2"
        print >>sys.stderr, "    It is recommended to use pypy instead of python"
        sys.exit(0)

    solution_number = sys.argv[1]
    orbit_number = int(sys.argv[2])
    nr = int(solution_number)

    #
    # The solutions in the file are separated by a blank line
    #
    solutions_txt = open("qSteinerSystems.txt",'r').read().strip().split('\n\n')
    print "Number of solutions in this file:", len(solutions_txt)

    if not (0 <= nr < len(solutions_txt)):
        print >>sys.stderr, ">>> Wrong solution number %d, must be between 0 and %d." % (nr, len(solutions_txt) - 1)
        sys.exit(1)
    #
    # Extract solutions from our set of solutions
    #
    solution_list = []
    for solution in [solution.split("\n")[1:] for solution in solutions_txt]:
        sol_vecs = []
        for line in solution:
            w = [[int(i) for i in v] for v in line.strip().split(' ')]
            sol_vecs.append(w)
        solution_list.append(sol_vecs)

    print "Solution #%s" %  solution_number
    sol = solution_list[nr]

    #
    #    Print solution
    #
    PrintOrbitRepresentatives(sol)

    print "\nSelect orbit %d" % orbit_number
    if not (0 <= orbit_number < len(sol)):
        print >>sys.stderr, ">>> Wrong orbit %d, number must be between 0 and %d." % (orbit_number, len(sol) - 1)
        sys.exit(1)
    IntersectionNumbers(sol, orbit_number)
